#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#
import random
import time
import os
import json

from netq_agent.cmd.base_wjh_agent import BaseWjhClient
from netq_lib.orm.redisdb.models import WjhDropAggregate, WjhL1DropAggregate

class DummyWjhClient(BaseWjhClient):

    def __init__(self, conf, logger):
        super(DummyWjhClient, self).__init__(logger)
        self._conf = conf
        self.simulate_wjh_data = self._conf.get('netq-agent', {}).get(
            "simulate-wjh", "off")

    def start_wjh_client(self):
        self._logger.info('Started DUMMY wjh agent, simulated-wjh flag: {}'.format(self.simulate_wjh_data))
        self.running = True

        # Collection every 60 seconds
        # Flush events every 15 seconds
        check_flush = 0
        flush_interval = 3
        dump_data = 0
        dump_data_interval = 12

        sleep_interval_seconds = 5
        while self.running:
            time.sleep(sleep_interval_seconds)
            dump_data += 1
            if dump_data == dump_data_interval:
                dump_data = 0
                """
                   If simulate_wjh_data is 'file', it reads the file periodically line by line, assuming
                   each line is a valid JSON object, that contains valid wjh drop information.
                   If simulate_wjh_data is 'random', it will generate random drops to the user periodically.
                   Else it will skip
                """
                if self.simulate_wjh_data == "random":
                    drops = generate_random_wjh_data(num_of_drops=5)
                    self.read_dummy_wjh_data(drops)
                elif self.simulate_wjh_data == "file":
                    data_file = '/etc/netq/.wjh.data'
                    if os.path.exists(data_file):
                        with open(data_file, 'r') as file:
                            drops = [json.loads(line) for line in file]
                        self.read_dummy_wjh_data(drops)
            check_flush += 1
            if check_flush == flush_interval:
                check_flush = 0
                self.flush_agg_cache()

    def read_dummy_wjh_data(self, drops):

        try:
            for drop_data in drops:
                if drop_data:
                    drop_type = drop_data.get('drop_type')
                    curr_timestamp = time.time()
                    first_timestamp = curr_timestamp - 60
                    last_timestamp = curr_timestamp
                    if drop_type in ['L2', 'Router', 'Tunnel', 'ACL', 'Buffer']:
                        self._logger.debug('{}'.format(drop_data))
                        wjh_drop = WjhDropAggregate()
                        wjh_drop.reason = drop_data.get('reason', '')
                        wjh_drop.drop_type = drop_type
                        wjh_drop.reason_id = drop_data.get('reason_id', 0)
                        wjh_drop.ingress_port = drop_data.get('ingress_port', '')
                        wjh_drop.egress_port = drop_data.get('egress_port', '')
                        wjh_drop.traffic_class = 0 if not drop_data.get('traffic_class') else int(drop_data.get('traffic_class'))
                        wjh_drop.acl_rule_id = 0 if not drop_data.get('acl_rule_id') else int(drop_data.get('acl_rule_id'))
                        wjh_drop.acl_bind_point = 0 if not drop_data.get('acl_bind_point') else int(drop_data.get('acl_bind_point'))
                        wjh_drop.acl_name = drop_data.get('acl_name', '')
                        wjh_drop.acl_rule = drop_data.get('acl_rule', '')
                        wjh_drop.smac = drop_data.get('smac', '')
                        wjh_drop.dmac = drop_data.get('dmac', '')
                        wjh_drop.proto = 0 if not drop_data.get('proto') else int(drop_data.get('proto'))
                        wjh_drop.sport = 0 if not drop_data.get('sport') else int(drop_data.get('sport'))
                        wjh_drop.dport = 0 if not drop_data.get('dport') else int(drop_data.get('dport'))
                        wjh_drop.sip = drop_data.get('sip', '')
                        wjh_drop.dip = drop_data.get('dip', '')
                        wjh_drop.first_timestamp = first_timestamp
                        wjh_drop.last_timestamp = last_timestamp
                        wjh_drop.agg_count = random.randint(1000, 10000)
                        wjh_drop.severity = drop_data.get('severity', '')
                        wjh_drop.traffic_class = drop_data.get('traffic_class', 0)
                        wjh_drop.original_latency = drop_data.get('original_latency', 0)
                        wjh_drop.original_occupancy = drop_data.get('original_occupancy', 0)
                        key = (wjh_drop.reason, wjh_drop.ingress_port, wjh_drop.sip, wjh_drop.dip,
                               wjh_drop.proto, wjh_drop.sport, wjh_drop.dport, wjh_drop.smac,
                               wjh_drop.dmac)
                        self.agg_cache[key] = wjh_drop
                        self._logger.debug('--- Others saved')
                    elif drop_type == "L1":
                        self._logger.debug('{}'.format(drop_type))
                        L1drop = WjhL1DropAggregate()
                        L1drop.ingress_port = drop_data.get('ingress_port', '')
                        L1drop.is_port_up = drop_data.get('is_port_up') == 'True' or 'False'
                        L1drop.port_down_reason = drop_data.get('port_down_reason', '')
                        L1drop.description = ''
                        if drop_data.get('state_change_count'):
                            L1drop.state_change_count = int(random.randint(1, 10))
                        if drop_data.get('symbol_error_count'):
                            L1drop.symbol_error_count = int(random.randint(1000, 10000))
                        if drop_data.get('crc_error_count'):
                            L1drop.crc_error_count = int(random.randint(1000, 100000))
                        L1drop.first_timestamp = first_timestamp
                        L1drop.last_timestamp = last_timestamp
                        L1drop.save()
                        self._logger.debug('--- L1 saved')

        except Exception as e:
            self._logger.error('Error parsing wjh data file: {}'.format(e))

def generate_random_wjh_data(num_of_drops):
        drops =  [
            # L1 drops
            {"drop_type": "L1", "description": "N/A", "ingress_port": "swp1", "is_port_up": True, "port_down_reason": "N/A", "symbol_error_count": 364},
            {"drop_type": "L1", "description": "N/A", "ingress_port": "swp1", "is_port_up": True, "port_down_reason": "N/A", "state_change_count": 2},
            {"drop_type": "L1", "description": "N/A", "ingress_port": "swp1", "is_port_up": True, "port_down_reason": "N/A", "crc_error_count": 80},

            # L2 drops
            {"drop_type": "L2", "agg_count": 556, "dip": "224.0.0.0", "dmac": "01:00:5e:61:f5:e7",
             "dport": 3300, "ingress_port": "swp1", "proto": 6,
             "reason": "Ingress VLAN filtering", "reason_id": 204,
             "severity": "Error", "sip": "16.0.0.2",
             "smac": "04:3f:72:db:fc:0d", "sport": 5001, "vlan": 20},
            {"drop_type": "L2", "agg_count": 536, "dip": "192.100.1.15",
             "dmac": "0c:42:a1:5a:5a:61", "dport": 5000,
             "ingress_port": "swp1", "proto": 6,
             "reason": "Ingress VLAN filtering", "reason_id": 204,
             "severity": "Error", "sip": "127.0.0.1",
             "smac": "04:3f:72:db:fc:0d", "sport": 5001, "vlan": 20},
            {"drop_type": "L2", "agg_count": 567, "dip": "0.0.0.0", "dmac": "1c:34:da:17:93:4c",
             "dport": 0, "ingress_port": "swp1", "proto": 0,
             "reason": "Source MAC is multicast", "reason_id": 209,
             "severity": "Error", "sip": "0.0.0.0", "smac": "01:00:5e:00:00:01",
             "sport": 0},
            {"drop_type": "L2", "agg_count": 566,
             "dip": "0.0.0.0", "dmac": "00:02:10:00:00:01",
             "dport": 0, "ingress_port": "swp1",
             "proto": 0, "reason": "Source MAC equals destination MAC", "reason_id": 210,
             "severity": "Error", "sip": "0.0.0.0", "smac": "00:02:10:00:00:01",
             "sport": 0},
            {"drop_type": "L2", "agg_count": 539, "dip": "192.100.1.18",
             "dmac": "ff:ff:ff:ff:ff:ff", "dport": 4000, "ingress_port": "swp1", "proto": 6,
             "reason": "Ingress VLAN filtering", "reason_id": 204,
             "severity": "Error", "sip": "192.100.1.18",
             "smac": "00:00:00:00:00:00", "sport": 5001, "vlan": 20},
            {"drop_type": "L2", "agg_count": 562,
             "dip": "192.100.1.15",
             "dmac": "1c:34:da:17:93:4c", "dport": 0, "ingress_port": "swp1",
             "proto": 0, "reason": "Ingress VLAN filtering", "reason_id": 204,
             "severity": "Error", "sip": "192.100.1.18",
             "smac": "04:3f:72:db:fc:0d", "sport": 0,
             "vlan": 100},
            {"drop_type": "L2", "agg_count": 536,
              "dip": "127.0.0.1", "dmac": "ff:ff:ff:ff:ff:ff",
             "dport": 5000, "ingress_port": "swp1", "proto": 17,
             "reason": "Ingress VLAN filtering", "reason_id": 204,
             "severity": "Error", "sip": "192.100.1.18",
             "smac": "00:00:00:00:00:00", "sport": 4001,
             "vlan": 20},
            {"drop_type": "L2", "agg_count": 557,
             "dip": "192.100.1.15",
             "dmac": "01:80:c2:00:00:01", "dport": 4000,
             "ingress_port": "swp1", "proto": 6,
             "reason": "Destination MAC is reserved (DMAC=01-80-C2-00-00-0x)",
             "reason_id": 202, "severity": "Error", "sip": "192.100.1.18",
             "smac": "04:3f:72:db:fc:0d", "sport": 5001},

            # Router drops
            {"drop_type": "Router", "dip": "20.20.20.20", "dmac": "b8:59:9f:a9:9f:18", "dport": 0,
              "ingress_port": "swp1", "proto": 61,
              "reason": "Checksum or IPver or IPv4 IHL too short",
              "reason_id": 314, "severity": "Error", "sip": "10.10.10.10",
              "smac": "00:00:00:00:00:00", "sport": 0},
            {"drop_type": "Router", "dip": "0.1.2.3", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 61,
              "reason": "IPv4 destination IP is local network (destination=0.0.0.0/8)",
              "reason_id": 318, "severity": "Error", "sip": "10.20.30.40",
              "smac": "00:00:00:00:00:00", "sport": 0},
            {"drop_type": "Router", "dip": "127.1.1.1", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 61,
              "reason": "Destination IP is loopback address", "reason_id": 309,
              "severity": "Error", "sip": "10.20.30.40",
              "smac": "00:00:00:00:00:00", "sport": 0},
            {"drop_type": "Router", "dip": "0.1.2.3", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 61,
              "reason": "IPv4 source IP is limited broadcast", "reason_id": 317,
              "severity": "Error", "sip": "255.255.255.255",
              "smac": "00:00:00:00:00:00", "sport": 0},
            {"drop_type": "Router", "dip": "224.0.0.1", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 61,
              "reason": "Multicast MAC mismatch", "reason_id": 315,
              "severity": "Error", "sip": "10.20.30.40",
              "smac": "00:00:00:00:00:00", "sport": 0,},
            {"drop_type": "Router", "dip": "40.30.20.10",
              "dmac": "b8:59:9f:a9:9f:18", "dport": 0, "proto": 61,
              "reason": "Source IP is in class E", "reason_id": 311,
              "severity": "Error", "sip": "240.0.0.1",
              "smac": "00:00:00:00:00:00", "sport": 0},
            {"drop_type": "Router", "dip": "2000::", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 59,
              "reason": "Source IP is loopback address", "reason_id": 312,
              "severity": "Error", "sip": "::1", "smac": "00:00:00:00:00:00",
              "sport": 0},
            {"drop_type": "Router", "dip": "2000::", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 59,
              "reason": "Source IP is multicast", "reason_id": 310,
              "severity": "Error", "sip": "ff00::1",
              "smac": "00:00:00:00:00:00", "sport": 0},
            {"drop_type": "Router", "dip": "2000::", "dmac": "b8:59:9f:a9:9f:18",
              "dport": 0, "ingress_port": "swp1", "proto": 59,
              "reason": "Source IP is unspecified", "reason_id": 313,
              "severity": "Error", "sip": "::", "smac": "00:00:00:00:00:00",
              "sport": 0},

            # Buffer drops
            {"agg_count": 1, "dip": "ff02::2", "dmac": "33:33:00:00:00:02", "dport": 0, "drop_type": "Buffer", "egress_port": "swp3", "ingress_port": "swp2",  "proto": 58,
             "reason": "Port TC Congestion Threshold Crossed", "reason_id": 505, "severity": "Warning", "sip": "fe80::6032:fdff:fe0f:e968",
             "smac": "62:32:fd:0f:e9:68", "sport": 0,  "traffic_class": 8, "original_latency": -1, "original_occupancy": -1},
            {"agg_count": 1, "dip": "11.1.1.2", "dmac": "31c:34:da:17:88:60", "dport": 0, "drop_type": "Buffer", "ingress_port": "swp2", "proto": 58,
             "reason": "Packet Latency Threshold Crossed", "reason_id": 506, "severity": "Warning", "sip": "12.1.1.2",
             "smac": "1c:34:da:17:93:60", "sport": 0, "traffic_class": 0, "original_latency": 0, "original_occupancy": 0},

            # ACL drops
            {"acl_bind_point": 0, "acl_name": "", "acl_rule": "",
             "acl_rule_id": 0, "agg_count": 201, "dip": "ff02::2", "dmac": "33:33:00:00:00:02",
             "dport": 0, "drop_type": "ACL", "egress_port": "swp3", "ingress_port": "swp2", "proto": 58, "reason": "Ingress port ACL",
             "reason_id": 601, "severity": "Notice", "sip": "fe80::6032:fdff:fe0f:e968", "smac": "62:32:fd:0f:e9:68", "sport": 0}
        ]
        random_numbers = random.sample(range(0, len(drops) - 1), num_of_drops)
        return [drops[idx] for idx in random_numbers]
