#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Lldp
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate, send_list_data, is_active_systemd_svc
from netq_agent.parsers.common.render import BaseRender


class LldpRender(BaseRender):
    def __init__(self):
        super(LldpRender, self).__init__()
        self.model = Lldp

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        '''
        Process LLDPD's lldpctl output and push appropriate LLDP data to DB
        '''
        retcode = ServiceStatus.OK
        fullstate = kwargs.get('fullstate')
        platform = kwargs.get('platform')

        if not is_active_systemd_svc('lldpd.service'):
            m_lldp = Lldp.memmodel()
            for entry in m_lldp.query.filter(hostname=HOSTNAME):
                old_lldp_ent = Lldp()
                old_lldp_ent.copy(entry)
                old_lldp_ent.delete()
            dummy = []
            if fullstate is True:
                save_fullstate(dummy, Lldp().__class__.__name__)
            return retcode

        try:
            # sometimes we get embedded double-quotes in json strings
            # that are added with backslash. these get removed
            # when using 'string-escape'. so we need to
            # double 'escape' the strings
            # we need to use 'unicode_escape' (python3) to handle Cisco
            # devices that split descriptions into multi-line outputs
            cmd_output = cmd_output.replace("\\\"", "\\\\\"")
            cmd_output = bytes(cmd_output, 'utf-8').decode("unicode_escape")
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            return ServiceStatus.FAIL

        # grab current list of lldp objects
        m_lldp = Lldp.memmodel()
        old_lldp_objs = []
        new_lldp_objs = []
        fullstate_lldp_objs = []
        if m_lldp:
            old_lldp_objs = (
                list(m_lldp.query.filter(hostname=HOSTNAME))
            )

        for values in list(json_out.values()):
            if isinstance(values, list):
                for value in values:
                    if not value.get('interface', None):
                        continue
                    if isinstance(value['interface'], list):
                        for lldp in value['interface']:
                            new_lldp_ent = self.parse_lldp(lldp, platform)
                            new_lldp_objs.append(new_lldp_ent)
                    else:
                        new_lldp_ent = self.parse_lldp(value['interface'], platform)
                        new_lldp_objs.append(new_lldp_ent)
            else:
                if not values.get('interface', None):
                    continue
                if isinstance(values['interface'], list):
                    for lldp in values['interface']:
                        new_lldp_ent = self.parse_lldp(lldp, platform)
                        new_lldp_objs.append(new_lldp_ent)
                else:
                    new_lldp_ent = self.parse_lldp(values['interface'], platform)
                    new_lldp_objs.append(new_lldp_ent)

        if fullstate is True:
            fullstate_lldp_objs = new_lldp_objs.copy()

        old_lldp_objs_copy = old_lldp_objs.copy()
        # remove current elements from old list
        for new_ele in new_lldp_objs[:]:
            del_new_ele = False
            for old_ele in old_lldp_objs[:]:
                if old_ele.equals(new_ele):
                    old_lldp_objs_copy.remove(old_ele)
                    old_ele_upd = Lldp()
                    old_ele_upd.copy(old_ele)
                    old_ele_upd.deleted = True
                    old_lldp_objs_copy.append(old_ele_upd)
                    del_new_ele = True
                elif old_ele.ifname == new_ele.ifname:
                    del_new_ele = True
            if del_new_ele:
                new_lldp_objs.remove(new_ele)
            else:
                if new_ele.peer_hostname != 'cumulus':
                    new_ele.save_record()

        # delete remaining stale entries
        for old_ele in old_lldp_objs_copy:
            if old_ele.deleted:
                continue
            old_lldp_ent = Lldp()
            old_lldp_ent.copy(old_ele)
            old_lldp_ent.delete()

        if fullstate is True:
            save_fullstate(fullstate_lldp_objs, Lldp().__class__.__name__)
        elif new_lldp_objs:
            send_list_data(new_lldp_objs, Lldp().__class__.__name__)

        return retcode

    def parse_capability(self, capability_list):
        capability = {}

        if not isinstance(capability_list, list):
            return capability

        for cap in capability_list:
            if cap['type'] == 'Bridge':
                capability['bridge'] = cap['enabled']
            elif cap['type'] == 'Router':
                capability['router'] = cap['enabled']
            elif cap['type'] == 'Station':
                capability['station'] = cap['enabled']

        return capability


    def parse_os_info(self, descr):
        words = descr.split()
        peer_os = ''
        peer_osv = ''

        if words[0] == 'Cumulus':
            peer_os = (' '.join(words[:2]))
            try:
                if "(" in words[3]:
                    peer_osv = (words[4].split(')')[1])
                else:
                    peer_osv = words[3]
            except Exception:
                peer_osv = (' '.join(words[3:4]))
        elif words[0] == 'Ubuntu':
            peer_os = words[0]
            peer_osv = words[1]
        elif words[0] == 'SONiC':
            peer_os = words[0]
            peer_info = dict()
            try:
                for info in descr.split(' - '):
                    [k, v] = info.split(':')
                    peer_info[k.strip()] = v.strip()
            except ValueError:
                pass    # continue, nothing to be done.
            peer_osv = peer_info.get('SONiC Software Version')
        else:
            peer_os = words[0]
            if "revision" in words:
                peer_osv = words[words.index("revision") + 1]
            elif "version" in words:
                peer_osv = words[words.index("version") + 1]
            else:
                peer_osv = "unknown"
        return (peer_os, peer_osv)


    def parse_host_lldp(self, lldp):
        lldpobj = Lldp()
        #
        # Walk through each interface
        #
        for ifname, lldp_info in list(lldp.items()):
            lldpobj.ifname = str(ifname)

            for key, value in list(lldp_info.items()):
                if key == 'chassis':
                    for peer, peer_info in list(value.items()):
                        if peer != 'id':
                            lldpobj.peer_hostname = str(peer)
                            for peer_attr, peer_attr_value in list(peer_info.items()):
                                if peer_attr == 'descr':
                                    peer_os, peer_osv = self.parse_os_info(peer_attr_value)
                                    lldpobj.lldp_peer_osv = peer_osv
                                    lldpobj.lldp_peer_os = peer_os
                                if peer_attr == 'capability':
                                    capability = self.parse_capability(peer_attr_value)
                                    lldpobj.lldp_peer_bridge = capability.get('bridge',
                                                                              False)
                                    lldpobj.lldp_peer_router = capability.get('router',
                                                                              False)
                                    lldpobj.lldp_peer_station = capability.get(
                                        'station', False)
                if key == 'port':
                    port_id = value.get('id', {})
                    lldpobj.peer_ifname = str(port_id.get('value'))

        lldpobj.save()
        return lldpobj
