#
# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import ctypes
import json
import subprocess
from datetime import datetime
from netq_lib.common.enums import ConfKw
from netq_lib.common import utils, configuration
from netq_lib.common.enums import ServiceStatus
from netq_lib.orm.redisdb.models import BERInfo
from netq_lib.common.utils import (
    subprocess_wrapper,
    is_active_systemd_svc,
    send_list_data,
    DEFAULT_TIMEOUT,
    check_cl_version_onwards
)
from netq_agent.parsers.common.render import BaseRender
from netq_agent.handlers.port_handler import PortHandler

SYSTEMCTL = '/bin/systemctl'

# Payload batch size to be sent to GRPC
BATCH_SIZE = 150

# Default values
NOT_AVAILABLE = 'N/A'
PORT_STATUS_DOWN = 'down'
PORT_STATUS_UP = 'Up'
MINUS_ONE = -1
DOUBLE_ZERO = 0.0
MINUS_ONE_STR = '-1'
EMPTY_STR = ''
UTF_ENCODING = 'utf-8'
UTF_ENCODING_REPLACE = 'replace'
M_CHAR = 'm'
STATUS_FIELD = 'status'
FEC_FIELD = 'fec'
UNKNOWN_VALUE = 'unknown'
PLATFORM_STR = 'platform'
RESULT_STR = 'result'
OUTPUT_STR = 'output'
VALUES_STR = 'values'
UNSUPPORTED = 'unsupported'
ZERO = '0'

# Fields to extract
EFFECTIVE_BER = 'effective_ber'
DEVICE_ID = 'device_id'
SYMBOL_ERRORS = 'symbol_errors'
DEVICE_FW_VERSION = 'device_fw_version'
MODULE_TEMP = 'module_temp'
ETHERNET_PROTOCOL_ACTIVE = 'ethernet_protocol_active'
PORT_NUMBER = 'port_number'
FC_ZERO_HIST = 'fc_zero_hist'
PHY_MANAGER_STATE = 'phy_manager_state'
CABLE_PN = 'cable_pn'
SPEED = 'speed_[gb/s]'
CABLE_SN = 'cable_sn'
ACTIVE_FEC = 'active_fec'
CABLE_TYPE = 'cable_type'
TIME_SINCE_LAST_CLEAR = 'time_since_last_clear_[min]'
CABLE_LENGTH = 'cable_length'
RAW_BER = 'raw_ber'
ADVANCED_STATUS_OPCODE = 'advanced_status_opcode'

IFNAME = 'ifname'

# Alternative Fields to extract
EFFECTIVE_BER_ALT = 'effective ber'
DEVICE_ID_ALT = 'device id'
DEVICE_FW_VERSION_ALT = 'fw version'
MODULE_TEMP_ALT = 'module start temp'
ETHERNET_PROTOCOL_ACTIVE_ALT = 'protocol'
CABLE_PN_ALT = 'cable pn'
SPEED_ALT = 'speed [gb/s]'
CABLE_SN_ALT = 'cable sn'
ACTIVE_FEC_ALT = 'active fec'
CABLE_TYPE_ALT = 'media'
TIME_SINCE_LAST_CLEAR_ALT = 'test time [min]'
CABLE_LENGTH_ALT = 'length [m]'
RAW_BER_ALT = 'total raw ber'


def to_double(arg, default=DOUBLE_ZERO):
    try:
        updated_arg = DOUBLE_ZERO if arg is None else arg
        casted_float = ctypes.c_double(float(updated_arg)).value if updated_arg.strip() else default
        return casted_float
    except ValueError:
        return DOUBLE_ZERO


def to_int(arg, default=MINUS_ONE_STR):
    try:
        updated_arg = MINUS_ONE if arg is None else arg
        casted_Int = ctypes.c_int(int(float(updated_arg))).value if updated_arg.strip() else default
        return casted_Int
    except ValueError:
        return MINUS_ONE


class DataBuffer:
    def __init__(self, threshold_size):
        self.threshold_size = threshold_size
        self.buffer = []
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)

    def add_data(self, data):
        self.buffer.append(data)
        if len(self.buffer) >= self.threshold_size:
            self.flush_buffer()

    def flush_buffer(self):
        if self.buffer:
            send_list_data(self.buffer, BERInfo().__class__.__name__)
            self.buffer = []


class BERRender(BaseRender):
    __logger = None

    def get_logger(self):
        if BERRender.__logger is None:
            __logger = self.__logger
        return BERRender.__logger

    def __init__(self):
        super(BERRender, self).__init__()
        self.model = BERInfo
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)
        self.buffer = None
        self.chip_type = None
        self.port_handler = PortHandler.get_instance(self.__logger, None)
        BERRender.__logger = self.__logger
        self.firmware_version = NOT_AVAILABLE

    # Function to get port details by oper-status
    def get_ports_by_status(self, data):
        up_ports = []
        down_ports = []
        for port_name, port_info in data.items():
            if port_info.get("type") != "swp":
                continue
            link_info = port_info.get("link", {})
            port_details = {
                "port": port_name,
                "speed": link_info.get("speed"),
                "oper-status": link_info.get("oper-status")
            }
            if link_info.get("oper-status") == "up":
                up_ports.append(port_details)
            elif link_info.get("oper-status") == "down":
                down_ports.append(port_details)
        return up_ports, down_ports

    def get_platform_firmware(self):
        if self.firmware_version != NOT_AVAILABLE:
            return self.firmware_version

        cmd = "nv show platform firmware -o=json"
        try:
            output = subprocess.check_output(cmd, timeout=30, shell=True).decode()
        except subprocess.CalledProcessError as e:
            self.__logger.info(str(e)) #pylint: disable=no-member
            return self.firmware_version

        data = json.loads(output)

        for key, value in data.items():
            if key.startswith("Spectrum"):
                self.firmware_version = value.get('actual-firmware')
                break

        return self.firmware_version

    def get_up_down_ports(self):
        cmd = "nv show interface -o=json"
        try:
            output = subprocess.check_output(cmd, timeout=30, shell=True).decode()
        except subprocess.CalledProcessError as e:
            self.__logger.debug(str(e)) #pylint: disable=no-member
            return ServiceStatus.FAIL

        data = json.loads(output)
        # Get ports with oper-status "up"
        oper_up_ports, oper_down_ports = self.get_ports_by_status(data)

        for port in oper_up_ports:
            self.__logger.debug("Ports with oper-status 'up': {} {} {}".format(port['port'], port['speed'], port['oper-status']))

        for port in oper_down_ports:
            self.__logger.debug("Ports with oper-status 'down': {} {} {}".format(port['port'], port['speed'], port['oper-status']))

        return oper_up_ports, oper_down_ports

    def generate_ber_info(self, up_ports):
        phy_detail_data = {}
        phy_diag_data = {}
        transceiver_data = {}
        interface_data = {}

        for port in up_ports:
            try:
                cmd = "nv show interface " + port['port'] + " link phy-detail -o=json"
                output = subprocess_wrapper(cmd.split(), timeout=DEFAULT_TIMEOUT)
                phy_detail_data = json.loads(output)

                cmd = "nv show interface " + port['port'] + " link phy-diag -o=json"
                output = subprocess_wrapper(cmd.split(), timeout=DEFAULT_TIMEOUT)
                phy_diag_data = json.loads(output)

                cmd = "nv show interface " + port['port'] + " -o=json"
                output = subprocess_wrapper(cmd.split(), timeout=DEFAULT_TIMEOUT)
                interface_data = json.loads(output)
                interface_pluggable = interface_data.get('pluggable', interface_data.get('transceiver', {}))

                if check_cl_version_onwards(major=5, minor=12):
                    cmd = "nv show interface " + port['port'] + " transceiver -o=json"
                    output = subprocess_wrapper(cmd.split(), timeout=DEFAULT_TIMEOUT)
                    transceiver_data = json.loads(output)

                self.generate_ber_payload(
                    phy_detail_data.get('effective-ber', NOT_AVAILABLE),
                    NOT_AVAILABLE,
                    to_int(phy_detail_data.get('symbol-errors', MINUS_ONE_STR)),
                    self.firmware_version,
                    to_int(interface_pluggable.get('temperature', MINUS_ONE_STR).split(" ")[0]),
                    port['speed'],
                    port['port'],
                    port['port'],
                    to_int(MINUS_ONE_STR),
                    to_int(MINUS_ONE_STR),
                    phy_diag_data.get('phy-manager-state', NOT_AVAILABLE),
                    interface_pluggable.get('vendor-pn', NOT_AVAILABLE),
                    port['speed'][:-1],
                    interface_pluggable.get('vendor-sn', NOT_AVAILABLE),
                    interface_data.get('link', {}).get('fec', NOT_AVAILABLE),
                    transceiver_data.get('cable-type', NOT_AVAILABLE),
                    to_double(phy_detail_data.get("time-since-last-clear-min", str(DOUBLE_ZERO))),
                    to_int(interface_pluggable.get('cable-length', MINUS_ONE_STR).replace(M_CHAR, EMPTY_STR)),
                    phy_detail_data.get("raw-ber", NOT_AVAILABLE),
                    NOT_AVAILABLE,
                    PORT_STATUS_UP)

            except Exception as e:
                if 'transceiver' in str(e):
                    self.__logger.debug('Unable to prepare BER response for port {} got exception {}'.format(port['port'], e))
                else:
                    self.__logger.warn('Unable to prepare BER response for port {} got exception {}'.format(port['port'], e))


    def generate_ber_payload(self, effective_ber, device_id, symbol_errors, device_fw_version, module_temp,
                             ethernet_protocol_active, ifname, port_number, local_port, fc_zero_hist,
                             phy_manager_state, cable_pn, speed, cable_sn, active_fec, cable_type,
                             time_since_last_clear,
                             cable_length, raw_ber, advanced_status_opcode, port_status):

        ber_info = BERInfo()
        ber_info.effective_ber = effective_ber
        ber_info.device_id = device_id
        ber_info.symbol_errors = symbol_errors if symbol_errors > 0 else 0
        ber_info.device_fw_version = device_fw_version
        ber_info.module_temp = module_temp
        ber_info.ethernet_protocol_active = ethernet_protocol_active
        ber_info.ifname = ifname
        ber_info.port_number = port_number
        ber_info.local_port = local_port
        ber_info.fc_zero_hist = fc_zero_hist
        ber_info.phy_manager_state = phy_manager_state
        ber_info.cable_pn = cable_pn
        ber_info.speed = speed
        ber_info.cable_sn = cable_sn
        ber_info.active_fec = active_fec
        ber_info.cable_type = cable_type
        ber_info.time_since_last_clear = time_since_last_clear
        ber_info.cable_length = cable_length
        ber_info.raw_ber = raw_ber
        ber_info.advanced_status_opcode = advanced_status_opcode
        ber_info.port_status = port_status.lower()

        self.buffer.add_data(ber_info)

    def generate_down_port_info(self, down_ports):
        for down_port in down_ports:
            self.generate_ber_payload(
                NOT_AVAILABLE, NOT_AVAILABLE, MINUS_ONE, NOT_AVAILABLE, MINUS_ONE, NOT_AVAILABLE,
                down_port['port'], down_port['port'], MINUS_ONE, MINUS_ONE, NOT_AVAILABLE, NOT_AVAILABLE, MINUS_ONE, NOT_AVAILABLE,
                NOT_AVAILABLE, NOT_AVAILABLE, DOUBLE_ZERO, MINUS_ONE, NOT_AVAILABLE, NOT_AVAILABLE, PORT_STATUS_DOWN
            )

        return ServiceStatus.OK

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        os.environ['HOME'] = "/root"
        datetimeNOW = datetime.now()
        self.__logger.debug('Going to collect BER info at {}'.format(datetimeNOW))

        try:
            if os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('nvued')

                if not is_active:
                    return ServiceStatus.IGNORE

        except Exception as e:
            self._logger.debug("nvued is not active")
            return ServiceStatus.IGNORE

        self.get_platform_firmware()
        up_ports, down_ports = self.get_up_down_ports()
        self.buffer = DataBuffer(threshold_size=BATCH_SIZE)

        self.generate_ber_info(up_ports)
        self.generate_down_port_info(down_ports)
        self.buffer.flush_buffer()
        return ServiceStatus.OK

def process_ber_info(cmd_output, svc_failed=False, **kwargs):
    ber_render = BERRender()

    return ber_render.parse(cmd_output, svc_failed, **kwargs)
