#
# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
from netq_agent.parsers.nvos.redis_db.client import NvosDBRender
from netq_agent.parsers.nvos.redis_db.nvos_db import NvosDB, NvosTable
from netq_lib.orm.redisdb.models import IbTransceiver


def _attr_state_db_trans_status_key_mapping():
    return {
        'temp_high_alarm_flag': 'temphighalarm_flag',
        'temp_low_alarm_flag': 'templowalarm_flag',
        'temp_high_warning_flag': 'temphighwarning_flag',
        'temp_low_warning_flag': 'temphighwarning_flag',
        'vcc_high_alarm_flag': 'vcchighalarm_flag',
        'vcc_low_alarm_flag': 'vcclowalarm_flag',
        'vcc_high_warning_flag': 'vcchighwarning_flag',
        'vcc_low_warning_flag': 'vcclowwarning_flag',
        'module_fw_fault': 'module_firmware_fault',
        'dp_fw_fault': 'datapath_firmware_fault',
    }


def _attr_state_db_trans_status_channel_key_mapping():
    return {
        'tx_fault': 'txfault',
        'tx_los': 'txlos_hostlane',
        'tx_cdr_lol': 'txcdrlol_hostlane',
        'tx_ad_eq_fault': 'txadeqfault_flag',
        'tx_power_high_alert': 'txpowerhighalarm_flag',
        'tx_power_low_alert': 'txpowerlowalarm_flag',
        'tx_power_high_warning': 'txpowerhighwarning_flag',
        'tx_power_low_warning': 'txpowerlowwarning_flag',
        'tx_bias_high_alert': 'txbiashighalarm_flag',
        'tx_bias_low_alert': 'txbiaslowalarm_flag',
        'tx_bias_high_warning': 'txbiashighwarning_flag',
        'tx_bias_low_warning': 'txbiaslowwarning_flag',
        'rx_los': 'rxlos',
        'rx_cdr_lol': 'rxcdrlol',
        'rx_power_high_alert': 'rxpowerhighalarm_flag',
        'rx_power_low_alert': 'rxpowerlowalarm_flag',
        'rx_power_high_warning': 'rxpowerhighwarning_flag',
        'rx_power_low_warning': 'rxpowerlowwarning_flag',
    }


def _attr_state_db_trans_dom_sensor_key_mapping():
    return {
        'laser_temperature': 'temperature',
        'supply_voltage': 'voltage'
    }


def extract_ifname_from_alias(ifname):
    match = re.match(r'(sw[A-Za-z]*\d+)', ifname)
    if match:
        return match.group()
    return ifname


TRANSCEIVER_STATUS_UNSET = '0'
TRANSCEIVER_STATUS_ACTIVE = '1'
NUMBER_OF_PHYSICAL_CHANNELS = 8


class TransceiverHandler(NvosDBRender):
    MODEL = IbTransceiver

    def __init__(self):
        super(TransceiverHandler, self).__init__()

    def db_sources(self):
        return {
            (NvosDB.CONFIG_DB, NvosTable.CONFIG_DB_IB_PORT): self.default_callback,
            (NvosDB.STATE_DB, NvosTable.STATE_DB_TRANSCEIVER_STATUS): self.default_callback,
            (NvosDB.STATE_DB, NvosTable.STATE_DB_TRANSCEIVER_DOM_SENSOR): self.default_callback,
        }

    def update_obj(self, obj, data, attr_map):
        if not data:
            return

        for attr_name, db_key in attr_map.items():
            try:
                attr_value = data.get(db_key)
                if attr_value is not None:
                    setattr(obj, attr_name, attr_value)
            except (ValueError, KeyError) as e:
                self._logger.debug(f'Failed to set attribute {attr_name} for object. Error: {e}')

    def get_dom_sensor_channel_list(self, data, prefix, suffix):
        # DOM sensor channel data is a string representing a float number, if applicable; otherwise, N/A.
        if not data:
            return []

        try:
            return [data.get(f"{prefix}{channel}{suffix}", 'N/A')
                    for channel in range(1, NUMBER_OF_PHYSICAL_CHANNELS + 1)]
        except Exception as e:
            self._logger.debug(f'Failed to set dom sensor channels list. Error: {e}')
            return []

    def update_dom_sensor_channels_data(self, obj, data):
        try:
            obj.output_power = self.get_dom_sensor_channel_list(data, 'tx', 'power')
            obj.input_power = self.get_dom_sensor_channel_list(data, 'rx', 'power')
            obj.laser_bias_current = self.get_dom_sensor_channel_list(data, 'tx', 'bias')
        except (ValueError, KeyError) as e:
            self._logger.debug(f'Failed to set DOM sensor channels data for object. Error: {e}')

    def update_channel_obj(self, obj, data, attr_map):
        # Update data per channel. Data should be stored in DB in the format db_key+<channel-index>, e.g., txfault1.
        # Each obj attribute in this method is stored as a list of boolean values, one per channel.
        # Assumption: db_key1-8 exists in the DB.

        if not data:
            for attr_name in attr_map:
                setattr(obj, attr_name, [])
            return

        for attr_name, db_key in attr_map.items():
            try:
                channel_list = [data.get(f'{db_key}{channel}', 'N/A') == 'True'
                                for channel in range(1, NUMBER_OF_PHYSICAL_CHANNELS + 1)]

                setattr(obj, attr_name, channel_list)
            except (ValueError, KeyError) as e:
                self._logger.debug(f'Failed to set attr {attr_name} for channel object. Error: {e}')

    def process_fullstate(self, data):
        trans_status = self.data_get(data, NvosDB.STATE_DB, NvosTable.STATE_DB_TRANSCEIVER_STATUS)
        dom_sensor = self.data_get(data, NvosDB.STATE_DB, NvosTable.STATE_DB_TRANSCEIVER_DOM_SENSOR)
        config_db_ip_port = self.data_get(data, NvosDB.CONFIG_DB, NvosTable.CONFIG_DB_IB_PORT)
        for ifname, status_json in trans_status.items():
            dom_sensor_info = dom_sensor.get(ifname)
            ib_transceiver_obj = IbTransceiver()
            alias = config_db_ip_port.get(ifname).get('alias', "")
            ib_transceiver_obj.name = extract_ifname_from_alias(alias)
            ib_transceiver_obj.module_oper_status = status_json.get('status', '0')
            ib_transceiver_obj.module_error_type = status_json.get('error', 'N/A')
            self.update_obj(ib_transceiver_obj, status_json, _attr_state_db_trans_status_key_mapping())
            self.update_channel_obj(ib_transceiver_obj, status_json,
                                    _attr_state_db_trans_status_channel_key_mapping())
            self.update_obj(ib_transceiver_obj, dom_sensor_info, _attr_state_db_trans_dom_sensor_key_mapping())
            self.update_dom_sensor_channels_data(ib_transceiver_obj, dom_sensor_info)
            ib_transceiver_obj.save()


def process_transceiver(*args, **kwargs):
    logger = kwargs.get('logger')
    render.process(logger)


render = TransceiverHandler()
